#include "prjlibs-c/standards.h"
#include <unistd.h>
#include <errno.h>
#include <sys/types.h>
#include <sys/stat.h>
#include <dirent.h>
#include <limits.h>
#include <stddef.h>
#include <ctype.h>

#include "skalibs/include/stddjb.h"
#include "prjlibs-c/constants.h"
#include "prjlibs-c/nsec.h"
#include "drop_root.h"
#include "die.h"
#include "timeout.h"

char const* PROG="posterity-pop";

static char buffer_space[8192];

typedef char const* ccp;

#define type_buf(type, name) typedef char name[(sizeof (type)*CHAR_BIT+2)/3+1]
type_buf(unsigned long, ulong_buf);
type_buf(uint64, uint64_buf);

static void fmt0_ulong(ulong_buf* buf, unsigned long x) {
  (*buf)[fmt_ulong(*buf, x)]='\0';
}

static void fmt0_uint64(uint64_buf* buf, uint64 x) {
  (*buf)[uint64_fmt(*buf, x)]='\0';
}

static void mustwrite(ccp text) {
  if (buffer_puts(buffer_1, text)<0) die_write();
}

static void respond(ccp r0, ccp r1, ccp r2, ccp r3) {
  mustwrite(r0);
  if (r1!=null) {
    mustwrite(r1);
    if (r2!=null) {
      mustwrite(r2);
      if (r3!=null) {
        mustwrite(r3);
      }
    }
  }
  mustwrite("\r\n");
  if (buffer_flush(buffer_1)<0) die_write();
}
#define RESPOND1(r0)             respond((r0), null, null, null)
#define RESPOND4(r0, r1, r2, r3) respond((r0), (r1), (r2), (r3))

enum state { state_auth, state_auth_user, state_trans };
static enum state state=state_auth;
#define ASSERT_STATE(assertion) \
  do if (state!=(assertion)) { \
    RESPOND1("-ERR that command is not valid in this state"); \
    return; \
  } while (0)

typedef struct message mesg_t;
struct message {
  off_t size;
  char name[1];
};
#define MESG_SIZE(namelen) (offsetof(mesg_t, name)+namelen)

typedef struct messagep mesgp_t;
struct messagep {
  mesg_t* mesg;
  mesgp_t* next;
  time_t mtime_secs;
  unsigned int mtime_nsecs;
};

static mesgp_t* mesgp_list=null;
static mesg_t** mesg_vec;
static unsigned long n_mesgs=0;
static stralloc dir=GEN_ALLOC_ZERO;
static ulong_buf pid_buf;
static uint64_buf size_buf;
static ccp ip="";
static char const ok_generic[]="+OK successful";

static mesg_t* get_mesg(ccp args) {
  unsigned long n;
  unsigned int i;
  i=scan_ulong(args, &n);
  if (i==0 || args[i]!='\0' || n==0 || n>n_mesgs) {
    RESPOND1("-ERR no such message");
    return null;
  }
  --n;
  return mesg_vec[n];
}

static void cmd_quit(char* args) {
  (void)args;
  RESPOND1("+OK quitting");
  _exit(0);
}

static void cmd_noop(char* args) {
  (void)args;
  ASSERT_STATE(state_trans);
  RESPOND1(ok_generic);
}
#define cmd_rset cmd_noop

static void cmd_user(char* args) {
  DIR* dirh;
  mesg_t** mesgv_tail;
  unsigned int size;
  uint64 total_size=0;
  ASSERT_STATE(state_auth);
  if (args[0]=='.' || args[str_chr(args, '/')]=='/' || chdir(args)!=0) {
    RESPOND1("-ERR no such mailbox");
    return;
  }
  if (stralloc_copys(&dir, args)==0 ||
      stralloc_cats(&dir, "/cur/")==0 ||
      stralloc_0(&dir)==0)
    strerr_die4sys(errstat, PROG, err_unable, err_alloc, ": ");
  if (chdir("cur")!=0) {
    RESPOND1("-ERR unable to access mailbox");
    strerr_die5sys(errstat, PROG, err_unable, err_chdirto, args, "/cur: ");
  }
  dirh=opendir(".");
  if (dirh==null) {
    RESPOND1("-ERR unable to open mailbox");
    strerr_die5sys(errstat, PROG, err_unable, err_open, args, "/cur: ");
  }
  for (;;) {
    struct stat statbuf;
    struct dirent* entry;
    mesgp_t* mesgp, ** mesgp_tail;
    mesg_t* mesg;
    errno=0;
    entry=readdir(dirh);
    if (entry==null) {
      if (errno==0) break;
      RESPOND1("-ERR unable to read mailbox");
      strerr_die5sys(errstat, PROG, err_unable, "read ", args, "/cur: ");
    }
    if (entry->d_name[0]=='.') continue;
    if (stat(entry->d_name, &statbuf)!=0) {
      RESPOND1("-ERR unable to stat message");
      strerr_die7sys(errstat, PROG, err_unable, "read ", args, "/cur/",
                     entry->d_name, ": ");
    }
    total_size+=statbuf.st_size;
    size=str_len(entry->d_name);
    if (++n_mesgs==0 || total_size<(uint64)statbuf.st_size ||
        entry->d_name[size]!='\0' || ++size==0) {
      RESPOND1("-ERR arithmetic overflow");
      die_overflow();
    }
    mesgp=(void*)alloc(sizeof *mesgp);
    mesg=(void*)alloc(MESG_SIZE(size));
    if (mesgp==null || mesg==null) {
      RESPOND1("-ERR unable to allocate memory");
      strerr_die4sys(errstat, PROG, err_unable, err_alloc, ": ");
    }
    mesg->size=statbuf.st_size;
    byte_copy(mesg->name, size, entry->d_name);
    mesgp->mesg=mesg;
    mesgp->mtime_secs=statbuf.st_mtime;
    mesgp->mtime_nsecs=mtime_nsec(&statbuf);
    for (mesgp_tail=&mesgp_list;; mesgp_tail=&(*mesgp_tail)->next) {
      if (*mesgp_tail==null) break;
      if ((*mesgp_tail)->mtime_secs <mesgp->mtime_secs)      continue;
      if ((*mesgp_tail)->mtime_secs >mesgp->mtime_secs)      break;
      if ((*mesgp_tail)->mtime_nsecs<mesgp->mtime_nsecs)     continue;
      if ((*mesgp_tail)->mtime_nsecs>mesgp->mtime_nsecs)     break;
      if (str_diff((*mesgp_tail)->mesg->name, mesg->name)<0) continue;
    }
    mesgp->next=*mesgp_tail;
    *mesgp_tail=mesgp;
  }
  closedir(dirh);
  fmt0_uint64(&size_buf, total_size);
  size=n_mesgs*sizeof (mesg_t*);
  if (size/n_mesgs!=sizeof (mesg_t*)) {
    RESPOND1("-ERR arithmetic overflow");
    die_overflow();
  }
  mesg_vec=(void*)alloc(size);
  if (mesg_vec==null) die_alloc();
  for (mesgv_tail=mesg_vec; mesgp_list!=null; ++mesgv_tail) {
    mesgp_t* next=mesgp_list->next;
    *mesgv_tail=mesgp_list->mesg;
    alloc_free(mesgp_list);
    mesgp_list=next;
  }
  state=state_auth_user;
  RESPOND1(ok_generic);
}

static void cmd_pass(char* args) {
  (void)args;
  ASSERT_STATE(state_auth_user);
  state=state_trans;
  RESPOND1(ok_generic);
}

static void cmd_stat(char* args) {
  ulong_buf ct_buf;
  (void)args;
  ASSERT_STATE(state_trans);
  fmt0_ulong(&ct_buf, n_mesgs);
  RESPOND4("+OK ", ct_buf, " ", size_buf);
}

static void list_line(ccp num_buf, mesg_t* mesg) {
  uint64_buf size_buf;
  fmt0_uint64(&size_buf, mesg->size);
  RESPOND4("+OK ", num_buf, " ", size_buf);
}

static void cmd_list(char* args) {
  ASSERT_STATE(state_trans);
  if (*args=='\0') {
    size_t i;
    for (i=0; i!=n_mesgs; ++i) {
      ulong_buf num_buf;
      fmt0_ulong(&num_buf, i+1);
      list_line(num_buf, mesg_vec[i]);
    }
    RESPOND1(".");
  } else {
    mesg_t* const mesg=get_mesg(args);
    if (mesg!=null) list_line(args, mesg);
  }
}

static void uidl_line(ccp num_buf, char* name) {
  unsigned int i=str_chr(name, ':'), colon=(name[i]==':');
  if (colon) name[i]='\0';
  RESPOND4("+OK ", num_buf, " ", name);
  if (colon) name[i]=':';
}

static void cmd_uidl(char* args) {
  ASSERT_STATE(state_trans);
  if (*args=='\0') {
    size_t i;
    for (i=0; i!=n_mesgs; ++i) {
      ulong_buf num_buf;
      fmt0_ulong(&num_buf, i+1);
      uidl_line(num_buf, mesg_vec[i]->name);
    }
    RESPOND1(".");
  } else {
    mesg_t* const mesg=get_mesg(args);
    if (mesg!=null) uidl_line(args, mesg->name);
  }
}

static void send_mesg(mesg_t const* mesg, int flag_top, unsigned long lines) {
  buffer input;
  int fd, flag_nl, flag_body=0;
  fd=open_read(mesg->name);
  strerr_warn6x(pid_buf, " ", ip, " read ", dir.s, mesg->name);
  if (fd<0) {
    RESPOND1("-ERR unable to open message");
    return;
  }
  RESPOND1("+OK message data follows");
  buffer_init(&input, &buffer_unixread, fd, buffer_space,
              sizeof(buffer_space));
  for (flag_nl=1;;) {
    char ch;
    int new_flag_nl, nread=buffer_GETC(&input, &ch);
    if (nread==0) break;
    if (nread<0) die_read(mesg->name);
    if (ch=='.' && flag_nl) mustwrite(".");
    new_flag_nl=(ch=='\n');
    if (new_flag_nl) {
      mustwrite("\r");
      if (flag_nl) flag_body=1;
    }
    flag_nl=new_flag_nl;
    if (buffer_PUTC(buffer_1, ch)<0) die_write();
    if (flag_top && flag_nl && flag_body && lines--==0) break;
  }
  close(fd);
  if (!flag_nl) RESPOND1("");
  RESPOND1(".");
}

static void cmd_top(char* args) {
  mesg_t const* mesg;
  unsigned long lines;
  unsigned int i;
  ASSERT_STATE(state_trans);
  i=str_chr(args, ' ');
  if (args[i]!=' ') {
    RESPOND1("-ERR bad syntax");
    return;
  }
  ((char*)args)[i]='\0';
  mesg=get_mesg(args);
  if (mesg==null) return;
  args+=i+1;
  i=scan_ulong(args, &lines);
  if (i==0 || args[i]!='\0') {
    RESPOND1("-ERR bad syntax");
    return;
  }
  send_mesg(mesg, 1, lines);
}

static void cmd_retr(char* args) {
  mesg_t* mesg;
  ASSERT_STATE(state_trans);
  mesg=get_mesg(args);
  if (mesg==null) return;
  send_mesg(mesg, 0, 0);
}

static void cmd_dele(char* args) {
  (void)args;
  ASSERT_STATE(state_trans);
  RESPOND1("-ERR this is a read-only service");
}

typedef struct {
  char name[5];
  void (*fn)(char*);
} command;

#define CMD(name) { #name, cmd_##name }
static command const commands[]={
  CMD(quit),
  CMD(user),
  CMD(pass),
  CMD(noop),
  CMD(list),
  CMD(uidl),
  CMD(stat),
  CMD(top),
  CMD(retr),
  CMD(dele),
  CMD(rset),
};
#undef CMD

static unsigned int const n_cmds=sizeof commands/sizeof commands[0];

int main(int argc, char** argv) {
  unsigned long n;
  unsigned int i;
  pid_t pid;
  stralloc sa=GEN_ALLOC_ZERO;

  (void)argv;
  if (argc>1)
    strerr_die3x(100, usage, "tcpserver 0 110 ", PROG);

  drop_root();
  timeout_setup();

  pid=getpid();
  fmt0_ulong(&pid_buf, pid);
  {
    ccp var=env_get("PROTO");
    if (var!=null) {
      if (stralloc_copys(&sa, var)==0 ||
          stralloc_cats(&sa, "REMOTEIP")==0 ||
          stralloc_0(&sa)==0)
        die_alloc();
      var=env_get(sa.s);
      if (var!=null) ip=var;
    }
  }

  RESPOND1("+OK posterity");

  for (;;) {
    char ch;
    if (stralloc_copys(&sa, "")==0) die_alloc();
    for (;;) {
      switch (buffer_GETC(buffer_0, &ch)) {
        case 0: _exit(0);
        case 1: break;
        default: die_read("network");
      }
      if (ch=='\n') break;
      ch=tolower(ch);
      if (stralloc_append(&sa, &ch)==0) die_alloc();
    }
    if (sa.s[sa.len-1]=='\r') --sa.len;
    if (byte_chr(sa.s, sa.len, '\0')!=sa.len) {
      RESPOND1("-ERR bad syntax");
      continue;
    }
    if (stralloc_0(&sa)==0) die_alloc();

    timeout_reset();

    n=byte_chr(sa.s, sa.len-1, ' ');

    for (i=0; i!=n_cmds; ++i)
      if (str_diffn(commands[i].name, sa.s, n)==0 &&
          n==str_len(commands[i].name))
        break;

    if (i==n_cmds)
      RESPOND1("-ERR unknown command");
    else
      commands[i].fn(sa.s+n+(sa.s[n]==' '));
  }
}
